/*
  This file contains the state space definition for Simple Sokoban.

  These are the only problem-specific (i.e., specific to Simple Sokoban)
  parts of the code; everything else is generic and can be used without
  change for other search problems.
*/

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Scanner;

public class SimpleSokobanStateSpace implements StateSpace {
    // Represents a grid cell that is part of the grid.
    private static class GridCell {
        public GridCell(int x, int y) {
            this.x = x;
            this.y = y;
        }

        public GridCell(GridCell other) {
            this.x = other.x;
            this.y = other.y;
        }

        public boolean equals(GridCell other) {
            return (this.x == other.x) && (this.y == other.y);
        }

        public String toString() {
            return "cell " + x + " / " + y;
        }

        public int x;
        public int y;
    }

    /*
      SimpleSokobanStates and SimpleSokobanActions are private since the
      search code cannot and should not look into the state. Since they
      are only accessed from this class, we can make the member
      variables public without violating encapsulation.
    */
    private static class SimpleSokobanState implements State {
        public GridCell agentPos;
        public GridCell[] boxPos;

        public SimpleSokobanState(GridCell agentPos, GridCell[] boxPos) {
            this.agentPos = agentPos;
            this.boxPos = boxPos;
        }

        @Override
        public boolean equals(Object other_) {
            if (!(other_ instanceof SimpleSokobanState)) {
                return false;
            }
            SimpleSokobanState other = (SimpleSokobanState)other_;
            if (this.boxPos.length != other.boxPos.length) {
                return false;
            }
            for (int index = 0; index < this.boxPos.length; ++index) {
                if (!this.boxPos[index].equals(other.boxPos[index])) {
                    return false;
                }
            }
            return this.agentPos.equals(other.agentPos);
        }

        // Tests for equality of two states, ignoring the position of
        // the agent (we could have chosen a different goal
        // representation, but using a SimpleSokobanState is a fairly
        // simple one).
        public boolean isGoalState(SimpleSokobanState goal) {
            assert (this.boxPos.length == goal.boxPos.length);
            for (int index = 0; index < this.boxPos.length; ++index) {
                if (!this.boxPos[index].equals(goal.boxPos[index])) {
                    return false;
                }
            }
            return true;
        }

        // Returns the index of the box at GridCell pos, and -1 if there
        // is no box.
        public int getBoxIndex(GridCell pos) {
            for (int index = 0; index < boxPos.length; ++index) {
                if (boxPos[index].equals(pos)) {
                    return index;
                }
            }
            return -1;
        }

        // Returns true if there is a box at GridCell pos.
        public boolean hasBoxAt(GridCell pos) {
            return getBoxIndex(pos) != -1;
        }

        public String toString() {
            String result = "agent: " + agentPos.toString() + "\n";
            for (int index = 0; index < boxPos.length; ++index) {
                result += "box " + index + ": " + boxPos[index].toString();
                if (index != boxPos.length -1) {
                    result += "\n";
                }
            }
            return result;
        }

        @Override
        public int hashCode() {
            return this.toString().hashCode();
        }
    }

    // The cost() function simplifies your implementation of action
    // costs, since you can add the cost directly to the actions
    // (instead of dealing with action costs in the cost() function of
    // the state space).
    private interface SimpleSokobanAction extends Action {
        public int cost();
    }

    // Represents a movement of the agent.
    private static class MoveAction implements SimpleSokobanAction {
        public GridCell from;
        public GridCell to;

        public MoveAction(GridCell from, GridCell to) {
            this.from = from;
            this.to = to;
        }

        public int cost() {
            return 1;
        }

        public String toString() {
            return "move(" + from.toString() + ", " + to.toString() + ")";
        }
    }

    // Represents the action where a box is pushed by one grid cell and
    // the agent moves along with it.
    private static class PushAction implements SimpleSokobanAction {
        public GridCell agentPos;
        public GridCell boxPos;
        public GridCell boxTargetPos;

        public PushAction(GridCell agentPos, GridCell boxPos, GridCell boxTargetPos) {
            this.agentPos = agentPos;
            this.boxPos = boxPos;
            this.boxTargetPos = boxTargetPos;
        }

        public int cost() {
            return 1;
        }

        public String toString() {
            return "push(" + agentPos.toString() + ", " + boxPos.toString() + ", " + boxTargetPos.toString() + ")";
        }
    }

    // The static parts of a task.
    public boolean[][] grid; // grid cell [i][j] is true if it can be
                             // entered, and false if it is a wall
    private int numberOfBoxes;
    private int gridWidth;
    private int gridHeight;
    
    private SimpleSokobanState initState;
    private SimpleSokobanState goalState;

    String heuristic;


    private SimpleSokobanStateSpace(boolean[][] grid, int numberOfBoxes,
                                    SimpleSokobanState initState,
                                    SimpleSokobanState goalState,
                                    String heuristic) {
        this.grid = grid;
        this.gridWidth = this.grid[0].length;
        this.gridHeight = this.grid.length;
        this.numberOfBoxes = numberOfBoxes;
        this.initState = initState;
        this.goalState = goalState;
        this.heuristic = heuristic;
        System.out.println("Instantiating problem instance with "
                           + numberOfBoxes + " boxes in a " + gridWidth
                           + "x" + gridHeight + " grid.");
    }

    /*
      The following five methods define the interface of state spaces used by
      the search code.

      We use the method names "init", "isGoal", "succ", "cost", and "h"
      to stay as close as possible to the names in the lecture slides.
      Without this constraint, it would be better to use more
      self-explanatory names like "getSuccessorStates" instead of
      "succ".
    */

    public State init() {
        // Just return the initial state that we stored.
        return initState;
    }

    public boolean isGoal(State s_) {
        SimpleSokobanState s = (SimpleSokobanState) s_;

        // Compare against the goal state that we stored.
        return s.isGoalState(goalState);
    }

    public ArrayList<StateActionPair> succ(State s_) {
        SimpleSokobanState s = (SimpleSokobanState) s_;
        GridCell pos = s.agentPos;
        // Create relevant neighbor cells
        GridCell north = new GridCell(pos.x, pos.y+1);
        GridCell south = new GridCell(pos.x, pos.y-1);
        GridCell west = new GridCell(pos.x-1, pos.y);
        GridCell east = new GridCell(pos.x+1, pos.y);
        GridCell north2 = new GridCell(pos.x, pos.y+2);
        GridCell south2 = new GridCell(pos.x, pos.y-2);
        GridCell west2 = new GridCell(pos.x-2, pos.y);
        GridCell east2 = new GridCell(pos.x+2, pos.y);

        ArrayList<StateActionPair> result = new ArrayList<StateActionPair>();

        /* We can only move to neighboring grid cells that aren't walls
         * and if there isn't a box
         */
        // move north
        if (gridCellIsAccessible(north) && !s.hasBoxAt(north)) {
            result.add(createMoveSuccessor(s, north));
        }

        // move south
        if (gridCellIsAccessible(south) && !s.hasBoxAt(south)) {
            result.add(createMoveSuccessor(s, south));
        }

        // move west
        if (gridCellIsAccessible(west) && !s.hasBoxAt(west)) {
            result.add(createMoveSuccessor(s, west));
        }

        // move east
        if (gridCellIsAccessible(east) && !s.hasBoxAt(east)) {
            result.add(createMoveSuccessor(s, east));
        }

         /* We can push only boxes that are on a neighboring cell and if
         the cell in the direction of the push is vacant
        */
        // push north
        if (s.hasBoxAt(north) && gridCellIsAccessible(north2) && !s.hasBoxAt(north2)) {
            result.add(createPushSuccessor(s, north, north2));
        }

        // push south
        if (s.hasBoxAt(south) && gridCellIsAccessible(south2) && !s.hasBoxAt(south2)) {
            result.add(createPushSuccessor(s, south, south2));
        }

        // push west
        if (s.hasBoxAt(west) && gridCellIsAccessible(west2) && !s.hasBoxAt(west2)) {
            result.add(createPushSuccessor(s, west, west2));
        }

        // push east
        if (s.hasBoxAt(east) && gridCellIsAccessible(east2) && !s.hasBoxAt(east2)) {
            result.add(createPushSuccessor(s, east, east2));
        }

        return result;
    }

    public int cost(Action a_) {
        // You don't have to change anything here, just adjust the
        // cost() functions of the MoveAction and PushAction classes
        SimpleSokobanAction a = (SimpleSokobanAction)a_;
        return a.cost();
    }

    public int h(State s_) {
        SimpleSokobanState s = (SimpleSokobanState) s_;
        switch (heuristic) {
        case "blind":
            // This is an exemplary implementation of the "blind"
            // heuristic, which returns 0 only for goal states.
            if (isGoal(s_)) {
                return 0;
            }
            return 1;
        case "h1":
            return h1(s);
        case "h2":
            return h2(s);
        case "h3":
            return h3(s);
        default:
            throw new IllegalArgumentException("Invalid heuristic: " + heuristic);
        }
    }

    private int h1(SimpleSokobanState s) {
        // Implement the heuristic h1 as described in Exercise 4.3.
        return -1;
    }

    private int h2(SimpleSokobanState s) {
        // Implement the heuristic h2 as described in Exercise 4.3.
        return -1;
    }

    private int h3(SimpleSokobanState s) {
        // Implement the heuristic h3 as described in Exercise 4.3.
        return -1;
    }

    // The following are helper functions to test applicability of an
    // action and to create successor states.
    private boolean gridCellIsAccessible(GridCell cell) {
        return (cell.x < gridWidth) && (cell.y < gridHeight) &&
            (cell.x >= 0) && (cell.y >= 0) && this.grid[cell.y][cell.x];
    }

    private StateActionPair createMoveSuccessor(SimpleSokobanState s, GridCell to) {
        GridCell[] boxPos = new GridCell[this.numberOfBoxes];
        for (int index = 0; index < boxPos.length; ++index) {
            boxPos[index] = s.boxPos[index];
        }
        return new StateActionPair(new SimpleSokobanState(to, boxPos), new MoveAction(s.agentPos, to));
    }

    private StateActionPair createPushSuccessor(SimpleSokobanState s, GridCell agentTo, GridCell boxTo) {
        assert(s.hasBoxAt(agentTo));
        int boxIndex = s.getBoxIndex(agentTo);
        GridCell[] boxPos = new GridCell[this.numberOfBoxes];
        for (int index = 0; index < boxPos.length; ++index) {
            if (index == boxIndex) {
                boxPos[index] = boxTo;
            } else {
                boxPos[index] = s.boxPos[index];
            }
        }
        return new StateActionPair(new SimpleSokobanState(agentTo, boxPos), new PushAction(s.agentPos, agentTo, boxTo));
    }

    /*
      The following methods instantiates the state space by reading the
      problem description from a file specified on the command line.

      Syntax of input: 
      - first line: grid width - grid height - number of boxes
      - second to (grid height +1)st line: 0/1 specifying if the grid cell can be entered
      - next line: initial agent position
      - final (number of boxes many) lines: initial and goal position of boxes
    */
    public static StateSpace buildFromCmdline(ArrayList<String> args) {
        if (args.size() != 2) {
            Errors.usageError("need one input file argument and a heuristic");
        }

        String filename = args.get(0);
        System.out.println("Reading input from file " + filename + "...");
        Scanner scanner;
        try {
            scanner = new Scanner(new File(filename));
        } catch (FileNotFoundException e) {
            Errors.fileError("input file not found: " + filename);
            scanner = new Scanner(""); // unreachable; silences compiler
        }

        int gridWidth = scanner.nextInt();
        int gridHeight = scanner.nextInt();
        int numBoxes = scanner.nextInt();

        if (numBoxes < 0 || gridWidth < 0 || gridHeight < 0) {
            Errors.fileError("invalid number of blocks or towers");
        }

        boolean[][] grid = scanGrid(gridWidth, gridHeight, scanner);
        GridCell initialAgentPos = new GridCell(scanner.nextInt(), scanner.nextInt());
        GridCell[] initialBoxPos = new GridCell[numBoxes];
        GridCell[] goalBoxPos = new GridCell[numBoxes];
        
        for (int index = 0; index < numBoxes; ++index) {
            initialBoxPos[index] = new GridCell(scanner.nextInt(), scanner.nextInt());
            goalBoxPos[index] = new GridCell(scanner.nextInt(), scanner.nextInt());
        }

        if (scanner.hasNext())
            Errors.fileError("expected end of file");
        
        scanner.close();

        GridCell[] boxes = {new GridCell(1,1)};
        SimpleSokobanState init = new SimpleSokobanState(initialAgentPos, initialBoxPos);

        // For the goal, the agent position doesn't matter so we can
        // give an arbitrary position.
        SimpleSokobanState goal = new SimpleSokobanState(new GridCell(0,0), goalBoxPos);

        String heuristic = args.get(1);
        if (!(heuristic.equals("blind") || heuristic.equals("h1") || 
              heuristic.equals("h2") || heuristic.equals("h3"))) {
            Errors.usageError("The heuristic parameter must be from the set {blind,h1,h2,h3}");
        }
        return new SimpleSokobanStateSpace(grid, numBoxes, init, goal, heuristic);
    }

    private static boolean[][] scanGrid(int gridWidth, int gridHeight, Scanner scanner) {
        boolean[][] result = new boolean[gridHeight][gridWidth];
        for (int i = 0; i < gridHeight; ++i) {
            for (int j = 0; j < gridWidth; ++j) {
                if (scanner.nextInt() == 1) {
                    result[i][j] = true;
                } else {
                    result[i][j] = false;
                }
            }
        }
        return result;
    }
}
